__version__ = "2.2.3"

def analyze(*args, **kwargs):
    from .analyzer import analyze as _analyze
    return _analyze(*args, **kwargs)

def debug_test():
    return "debug-ok"

__all__ = ["analyze", "debug_test", "__version__"]



#!/usr/bin/env python3
import ast
import sys
import json
import logging
from pathlib import Path
from collections import defaultdict
from skylos.visitor import Visitor
from skylos.constants import ( PENALTIES, AUTO_CALLED )
from skylos.visitors.test_aware import TestAwareVisitor
from skylos.rules.secrets import scan_ctx as _secrets_scan_ctx
import os
import traceback
from skylos.visitors.framework_aware import FrameworkAwareVisitor, detect_framework_usage

logging.basicConfig(level=logging.INFO,format='%(asctime)s - %(levelname)s - %(message)s')
logger=logging.getLogger('Skylos')

class Skylos:
    def __init__(self):
        self.defs={}
        self.refs=[]
        self.dynamic=set()
        self.exports=defaultdict(set)

    def _module(self,root,f):
        p = list(f.relative_to(root).parts)
        if p[-1].endswith(".py"):
            p[-1] = p[-1][:-3]
        if p[-1] == "__init__":
            p.pop()
        return ".".join(p)
    
    def _should_exclude_file(self, file_path, root_path, exclude_folders):
        if not exclude_folders:
            return False
            
        try:
            rel_path = file_path.relative_to(root_path)
        except ValueError:
            return False
        
        path_parts = rel_path.parts
        
        for exclude_folder in exclude_folders:
            if "*" in exclude_folder:
                for part in path_parts:
                    if part.endswith(exclude_folder.replace("*", "")):
                        return True
            else:
                if exclude_folder in path_parts:
                    return True
        
        return False
    
    def _get_python_files(self, path, exclude_folders=None):
        p = Path(path).resolve()
        
        if p.is_file():
            return [p], p.parent
        
        root = p
        all_files = list(p.glob("**/*.py"))
        
        if exclude_folders:
            filtered_files = []
            excluded_count = 0
            
            for file_path in all_files:
                if self._should_exclude_file(file_path, root, exclude_folders):
                    excluded_count += 1
                    continue
                filtered_files.append(file_path)
            
            if excluded_count > 0:
                logger.info(f"Excluded {excluded_count} files from analysis")
            
            return filtered_files, root
        
        return all_files, root
    
    def _mark_exports(self):
        for name, definition in self.defs.items():
            if definition.in_init and not definition.simple_name.startswith('_'):
                definition.is_exported = True
        
        for mod, export_names in self.exports.items():
            for name in export_names:
                for def_name, def_obj in self.defs.items():
                    if (def_name.startswith(f"{mod}.") and 
                        def_obj.simple_name == name and
                        def_obj.type != "import"):
                        def_obj.is_exported = True

    def _mark_refs(self):
        import_to_original = {}
        for name, def_obj in self.defs.items():
            if def_obj.type == "import":
                import_name = name.split('.')[-1]
                
                for def_name, orig_def in self.defs.items():
                    if (orig_def.type != "import" and 
                        orig_def.simple_name == import_name and
                        def_name != name):
                        import_to_original[name] = def_name
                        break

        simple_name_lookup = defaultdict(list)
        for definition in self.defs.values():
            simple_name_lookup[definition.simple_name].append(definition)
        
        for ref, _ in self.refs:
            if ref in self.defs:
                self.defs[ref].references += 1
                if ref in import_to_original:
                    original = import_to_original[ref]
                    self.defs[original].references += 1
                continue

            simple = ref.split('.')[-1]
            ref_mod = ref.rsplit(".", 1)[0]
            candidates = simple_name_lookup.get(simple, [])

            if ref_mod:
                filtered = []
                for d in candidates:
                    if d.name.startswith(ref_mod + ".") and d.type != "import":
                        filtered.append(d)
                candidates = filtered
            else:
                filtered = []
                for d in candidates:
                    if d.type != "import":
                        filtered.append(d)
                candidates = filtered

            if len(candidates) == 1:
                candidates[0].references += 1

        for module_name in self.dynamic:
            for def_name, def_obj in self.defs.items():
                if def_obj.name.startswith(f"{module_name}."):
                    if def_obj.type in ("function", "method") and not def_obj.simple_name.startswith("_"):
                        def_obj.references += 1
    
    def _get_base_classes(self, class_name):
        if class_name not in self.defs:
            return []
        
        class_def = self.defs[class_name]
        
        if hasattr(class_def, 'base_classes'):
            return class_def.base_classes
        
        return []
    
    def _apply_penalties(self, def_obj, visitor, framework):
        confidence=100

        if getattr(visitor, "ignore_lines", None) and def_obj.line in visitor.ignore_lines:
            def_obj.confidence = 0
            return

        if def_obj.type == "variable" and def_obj.simple_name == "_":
            def_obj.confidence = 0
            return

        if def_obj.simple_name.startswith("_") and not def_obj.simple_name.startswith("__"):
            confidence -= PENALTIES["private_name"] 

        if def_obj.simple_name.startswith("__") and def_obj.simple_name.endswith("__"):
            confidence -= PENALTIES["dunder_or_magic"]

        if def_obj.in_init and def_obj.type in ("function", "class"):
            confidence -= PENALTIES["in_init_file"]
            
        if def_obj.name.split(".")[0] in self.dynamic:
            confidence -= PENALTIES["dynamic_module"]

        if visitor.is_test_file or def_obj.line in visitor.test_decorated_lines:
            confidence -= PENALTIES["test_related"]

        if def_obj.type == "variable" and getattr(framework, "dataclass_fields", None):
            if def_obj.name in framework.dataclass_fields:
                def_obj.confidence = 0
                return
            
        if def_obj.type == "variable":
            fr = getattr(framework, "first_read_lineno", {}).get(def_obj.name)
            if fr is not None and fr >= def_obj.line:
                def_obj.confidence = 0
                return
        
        framework_confidence = detect_framework_usage(def_obj, visitor=framework)
        if framework_confidence is not None:
            confidence = min(confidence, framework_confidence)

        if (def_obj.simple_name.startswith("__") and def_obj.simple_name.endswith("__")):
            confidence = 0

        if def_obj.type == "parameter":
            if def_obj.simple_name in ("self", "cls"):
                confidence = 0
            elif "." in def_obj.name:
                method_name = def_obj.name.split(".")[-2]
                if method_name.startswith("__") and method_name.endswith("__"):
                    confidence = 0
        
        if visitor.is_test_file or def_obj.line in visitor.test_decorated_lines:
            confidence = 0
        
        if (def_obj.type == "import" and def_obj.name.startswith("__future__.") and
            def_obj.simple_name in ("annotations", "absolute_import", "division", 
                                "print_function", "unicode_literals", "generator_stop")):
            confidence = 0
        
        def_obj.confidence = max(confidence, 0)

    def _apply_heuristics(self):
        class_methods = defaultdict(list)
        for definition in self.defs.values():
            if definition.type in ("method", "function") and "." in definition.name:
                cls = definition.name.rsplit(".", 1)[0]
                if cls in self.defs and self.defs[cls].type == "class":
                    class_methods[cls].append(definition)
        
        for cls, methods in class_methods.items():
            if self.defs[cls].references > 0:
                for method in methods:
                    if method.simple_name in AUTO_CALLED:
                        method.references += 1
                    
                    if (method.simple_name.startswith("visit_") or
                        method.simple_name.startswith("leave_") or
                        method.simple_name.startswith("transform_")):
                        method.references += 1

                    if method.simple_name == "format" and cls.endswith("Formatter"):
                        method.references += 1

    def analyze(self, path, thr=60, exclude_folders= None, enable_secrets = False):
        files, root = self._get_python_files(path, exclude_folders)
        
        if not files:
            logger.warning(f"No Python files found in {path}")
            return json.dumps({
                "unused_functions": [], 
                "unused_imports": [], 
                "unused_classes": [],
                "unused_variables": [],
                "unused_parameters": [],
                 "analysis_summary": {
                    "total_files": 0,
                    "excluded_folders": exclude_folders if exclude_folders else []
                }
            })
        
        logger.info(f"Analyzing {len(files)} Python files...")
        
        modmap = {}
        for f in files:
            modmap[f] = self._module(root, f)
        
        all_secrets = []
        for file in files:
            mod = modmap[file]
            defs, refs, dyn, exports, test_flags, framework_flags = proc_file(file, mod)

            for definition in defs:
                self._apply_penalties(definition, test_flags, framework_flags) 
                self.defs[definition.name] = definition

            self.refs.extend(refs)
            self.dynamic.update(dyn)
            self.exports[mod].update(exports)

            if enable_secrets and _secrets_scan_ctx is not None:
                try:
                    src_lines = Path(file).read_text(encoding="utf-8", errors="ignore").splitlines(True)
                    ctx = {"relpath": str(file), "lines": src_lines, "tree": None}
                    findings = list(_secrets_scan_ctx(ctx))
                    if findings:
                        all_secrets.extend(findings)
                except Exception:
                    pass

        self._mark_refs()
        self._apply_heuristics() 
        self._mark_exports()

        shown = 0

        def def_sort_key(d):
            return (d.type, d.name)

        for d in sorted(self.defs.values(), key=def_sort_key):
            if shown >= 50:
                break
            print(f" type={d.type} refs={d.references} conf={d.confidence} exported={d.is_exported} line={d.line} name={d.name}")
            shown += 1

        unused = []
        for definition in self.defs.values():
                if definition.references == 0 and not definition.is_exported and definition.confidence > 0 and definition.confidence >= thr:
                    unused.append(definition.to_dict())

        result = {
            "unused_functions": [], 
            "unused_imports": [], 
            "unused_classes": [],
            "unused_variables": [],
            "unused_parameters": [],
            "analysis_summary": {
                "total_files": len(files),
                "excluded_folders": exclude_folders or [],
            }
        }

        if enable_secrets and all_secrets:
            result["secrets"] = all_secrets
        
        for u in unused:
            if u["type"] in ("function", "method"):
                result["unused_functions"].append(u)
            elif u["type"] == "import":
                result["unused_imports"].append(u)
            elif u["type"] == "class": 
                result["unused_classes"].append(u)
            elif u["type"] == "variable":
                result["unused_variables"].append(u)
            elif u["type"] == "parameter":
                result["unused_parameters"].append(u)
                
        return json.dumps(result, indent=2)

def proc_file(file_or_args, mod=None):
    if mod is None and isinstance(file_or_args, tuple):
        file, mod = file_or_args 
    else:
        file = file_or_args 

    try:
        source = Path(file).read_text(encoding="utf-8")
        ignore_lines = {i for i, line in enumerate(source.splitlines(), start=1)
                        if "pragma: no skylos" in line}
        tree = ast.parse(source)

        tv = TestAwareVisitor(filename=file)
        tv.visit(tree)
        tv.ignore_lines = ignore_lines

        fv = FrameworkAwareVisitor(filename=file)
        fv.visit(tree)
        fv.finalize()
        v  = Visitor(mod, file)
        v.visit(tree)

        fv.dataclass_fields = getattr(v, "dataclass_fields", set())
        fv.first_read_lineno = getattr(v, "first_read_lineno", {})

        return v.defs, v.refs, v.dyn, v.exports, tv, fv
    
    except Exception as e:
        logger.error(f"{file}: {e}")
        if os.getenv("SKYLOS_DEBUG"):
            logger.error(traceback.format_exc())
        dummy_visitor = TestAwareVisitor(filename=file)
        dummy_visitor.ignore_lines = set()
        dummy_framework_visitor = FrameworkAwareVisitor(filename=file)

        return [], [], set(), set(), dummy_visitor, dummy_framework_visitor

def analyze(path, conf=60, exclude_folders=None, enable_secrets=False):
    return Skylos().analyze(path,conf, exclude_folders, enable_secrets)

if __name__ == "__main__":
    if len(sys.argv)>1:
        p = sys.argv[1]
        confidence = int(sys.argv[2]) if len(sys.argv) >2 else 60
        result = analyze(p,confidence)
        
        data = json.loads(result)
        print("\n Python Static Analysis Results")
        print("===================================\n")
        
        total_items = 0
        for key, items in data.items():
            if key.startswith("unused_") and isinstance(items, list):
                total_items += len(items)
        
        print("Summary:")
        if data["unused_functions"]:
            print(f" * Unreachable functions: {len(data['unused_functions'])}")
        if data["unused_imports"]:
            print(f" * Unused imports: {len(data['unused_imports'])}")
        if data["unused_classes"]:
            print(f" * Unused classes: {len(data['unused_classes'])}")
        if data["unused_variables"]:
            print(f" * Unused variables: {len(data['unused_variables'])}")
        
        if data["unused_functions"]:
            print("\n - Unreachable Functions")
            print("=======================")
            for i, func in enumerate(data["unused_functions"], 1):
                print(f" {i}. {func['name']}")
                print(f"    └─ {func['file']}:{func['line']}")
        
        if data["unused_imports"]:
            print("\n - Unused Imports")
            print("================")
            for i, imp in enumerate(data["unused_imports"], 1):
                print(f" {i}. {imp['simple_name']}")
                print(f"    └─ {imp['file']}:{imp['line']}")
        
        if data["unused_classes"]:
            print("\n - Unused Classes")
            print("=================")
            for i, cls in enumerate(data["unused_classes"], 1):
                print(f" {i}. {cls['name']}")
                print(f"    └─ {cls['file']}:{cls['line']}")
                
        if data["unused_variables"]:
            print("\n - Unused Variables")
            print("==================")
            for i, var in enumerate(data["unused_variables"], 1):
                print(f" {i}. {var['name']}")
                print(f"    └─ {var['file']}:{var['line']}")
        
        print("\n" + "─" * 50)
        print(f"Found {total_items} dead code items. Add this badge to your README:")
        print(f"```markdown")
        print(f"![Dead Code: {total_items}](https://img.shields.io/badge/Dead_Code-{total_items}_detected-orange?logo=codacy&logoColor=red)")
        print(f"```")
        
        print("\nNext steps:")
        print("  * Use --interactive to select specific items to remove")
        print("  * Use --dry-run to preview changes before applying them")
    else:
        print("Usage: python Skylos.py <path> [confidence_threshold]")


#!/usr/bin/env python3
import ast
from pathlib import Path
import re

PYTHON_BUILTINS={"print", "len", "str", "int", "float", "list", "dict", "set", "tuple", "range", "open", "reversed", 
                 "super", "object", "type", "enumerate", "zip", "map", "filter", "sorted", "sum", "min", 
                "next", "iter", "bytes", "bytearray", "format", "round", "abs", "complex", "hash", "id", "bool", "callable", 
                "getattr", "max", "all", "any", "setattr", "hasattr", "isinstance", "globals", "locals", 
                "vars", "dir" ,"property", "classmethod", "staticmethod"}
DYNAMIC_PATTERNS={"getattr", "globals", "eval", "exec"}

## "🥚" hi :) 

class Definition:
    
    def __init__(self, name, t, filename, line):
        self.name = name
        self.type = t
        self.filename = filename
        self.line = line
        self.simple_name = name.split('.')[-1]
        self.confidence = 100
        self.references = 0
        self.is_exported = False
        self.in_init = "__init__.py" in str(filename)
    
    def to_dict(self):
        if self.type == "method" and "." in self.name:
            parts = self.name.split(".")
            if len(parts) >= 3:
                output_name = ".".join(parts[-2:])
            else:
                output_name = self.name
        else:
            output_name = self.simple_name
            
        return{
            "name": output_name,
            "full_name": self.name,
            "simple_name": self.simple_name,
            "type": self.type,
            "file": str(self.filename),
            "basename": Path(self.filename).name,
            "line": self.line,
            "confidence": self.confidence,
            "references": self.references
        }

class Visitor(ast.NodeVisitor):
    def __init__(self,mod,file):
        self.mod= mod
        self.file= file
        self.defs= []
        self.refs= []
        self.cls= None
        self.alias= {}
        self.dyn= set()
        self.exports= set()
        self.current_function_scope= []
        self.current_function_params= []
        self.local_var_maps = []
        self.in_cst_class = 0
        self.local_type_maps = []
        self._dataclass_stack = [] 
        self.dataclass_fields = set()
        self.first_read_lineno = {} 

    def add_def(self, name, t, line):
        found = False
        for d in self.defs:
            if d.name == name:
                found = True
                break
        if not found:
            self.defs.append(Definition(name, t, self.file, line))

    def add_ref(self, name):
        self.refs.append((name, self.file))

    def qual(self, name):
        if name in self.alias:
            return self.alias[name]
        if name in PYTHON_BUILTINS:
            if self.mod:
                mod_candidate = f"{self.mod}.{name}"
            else:
                mod_candidate = name
            if any(d.name == mod_candidate for d in self.defs):
                return mod_candidate
            
        if self.mod:
            return f"{self.mod}.{name}"
        else:
            return name
    
    def visit_Global(self, node):
        if self.current_function_scope and self.local_var_maps:
            for name in node.names:
                self.local_var_maps[-1][name] = f"{self.mod}.{name}"
        return

    def visit_annotation(self, node):
        if node is not None:
            if isinstance(node, ast.Constant) and isinstance(node.value, str):
                self.visit_string_annotation(node.value)
            elif hasattr(node, 's') and isinstance(node.s, str):
                self.visit_string_annotation(node.s)
            else:
                self.visit(node)

    def visit_string_annotation(self, annotation_str):
        if not isinstance(annotation_str, str):
            return

        try:
            parsed = ast.parse(annotation_str, mode="eval")
            self.visit(parsed.body)
        except:
            for tok in re.findall(r"[A-Za-z_][A-Za-z0-9_]*", annotation_str):
                self.add_ref(tok)

    def visit_Import(self, node):
        for a in node.names:
            full= a.name
            self.alias[a.asname or a.name.split(".")[-1]]= full
            self.add_def(full,"import",node.lineno)

    def visit_ImportFrom(self, node):
        if node.module is None:
            return
        for a in node.names:
            if a.name == "*":
                continue
            base = node.module
            if node.level:
                parts = self.mod.split(".")
                base = ".".join(parts[:-node.level]) + (f".{node.module}" if node.module else "")
            
            full = f"{base}.{a.name}"
            if a.asname:
                self.alias[a.asname] = full
                self.add_def(full, "import", node.lineno)
            else:
                self.alias[a.name] = full
                self.add_def(full, "import", node.lineno)
            
    def visit_arguments(self, args):
        for arg in args.args:
            self.visit_annotation(arg.annotation)
        for arg in args.posonlyargs:
            self.visit_annotation(arg.annotation)
            
        for arg in args.kwonlyargs:
            self.visit_annotation(arg.annotation)
        if args.vararg:
            self.visit_annotation(args.vararg.annotation)
        if args.kwarg:
            self.visit_annotation(args.kwarg.annotation)
        for default in args.defaults:
            self.visit(default)
        for default in args.kw_defaults:
            if default:
                self.visit(default)

    def visit_FunctionDef(self,node):
        outer_scope_prefix = '.'.join(self.current_function_scope) + '.' if self.current_function_scope else ''
        
        if self.cls:
            name_parts= [self.mod, self.cls, outer_scope_prefix + node.name]
        else:
            name_parts= [self.mod, outer_scope_prefix + node.name]
        
        qualified_name= ".".join(filter(None, name_parts))

        if self.cls:
            def_type = "method"
        else:
            def_type = "function"
        self.add_def(qualified_name, def_type, node.lineno)
        
        self.current_function_scope.append(node.name)
        self.local_var_maps.append({})
        self.local_type_maps.append({})

        old_params = self.current_function_params
        self.current_function_params = []
        
        for d_node in node.decorator_list:
            self.visit(d_node)
        
        for arg in node.args.args:
            param_name = f"{qualified_name}.{arg.arg}"
            self.add_def(param_name, "parameter", node.lineno)
            self.current_function_params.append((arg.arg, param_name))
        
        self.visit_arguments(node.args)
        self.visit_annotation(node.returns)
        
        for stmt in node.body:
            self.visit(stmt)
            
        self.current_function_scope.pop()
        self.current_function_params = old_params
        self.local_var_maps.pop()
        self.local_type_maps.pop()
        
    visit_AsyncFunctionDef= visit_FunctionDef

    def visit_ClassDef(self, node):
        cname =f"{self.mod}.{node.name}"
        self.add_def(cname, "class",node.lineno)
        
        is_cst = False
        is_dc  = False

        for base in node.bases:
            base_name = ""

            if isinstance(base, ast.Attribute):
                base_name = base.attr
                
            elif isinstance(base, ast.Name):
                base_name = base.id
            self.visit(base)

            if base_name in {"CSTTransformer", "CSTVisitor"}:
                is_cst = True
                
        for keyword in node.keywords:
            self.visit(keyword.value)

        for decorator in node.decorator_list:
            def _is_dc(dec):
                if isinstance(dec, ast.Call):
                    target = dec.func
                else:
                    target = dec

                if isinstance(target, ast.Name):
                    return target.id == "dataclass"
                
                if isinstance(target, ast.Attribute):
                    return target.attr == "dataclass"
                
                return False
            
            if _is_dc(decorator):
                is_dc = True
            self.visit(decorator)
        
        prev= self.cls
        if is_cst:
            self.in_cst_class += 1

        self.cls= node.name
        self._dataclass_stack.append(is_dc)
        for b in node.body:
            self.visit(b)

        self.cls= prev
        self._dataclass_stack.pop()
    
        if is_cst:
            self.in_cst_class -= 1

    def visit_AnnAssign(self, node):
        self.visit_annotation(node.annotation)
        if node.value:
            self.visit(node.value)
    
        def _define(t):
            if isinstance(t, ast.Name):
                name_simple = t.id
                scope_parts = [self.mod]
                if self.cls: scope_parts.append(self.cls)

                if self.current_function_scope: scope_parts.extend(self.current_function_scope)
                prefix = '.'.join(filter(None, scope_parts))
                if prefix:
                    var_name = f"{prefix}.{name_simple}"
                else:
                    var_name = name_simple

                self.add_def(var_name, "variable", t.lineno)
                if (self._dataclass_stack and self._dataclass_stack[-1] 
                    and self.cls 
                    and not self.current_function_scope):
                    self.dataclass_fields.add(var_name)

                if self.current_function_scope and self.local_var_maps:
                    self.local_var_maps[-1][name_simple] = var_name

            elif isinstance(t, (ast.Tuple, ast.List)):

                for elt in t.elts:
                    _define(elt)
        _define(node.target)

    def visit_AugAssign(self, node):
        if isinstance(node.target, ast.Name):
            nm = node.target.id
            if (self.current_function_scope and 
                self.local_var_maps and 
                self.local_var_maps and 
                nm in self.local_var_maps[-1]):

                # self.add_ref(self.local_var_maps[-1][nm])
                fq = self.local_var_maps[-1][nm]
                self.add_ref(fq)
                var_name = fq

            else:
                self.add_ref(self.qual(nm))
                scope_parts = [self.mod]
                if self.cls: 
                    scope_parts.append(self.cls)

                if self.current_function_scope: 
                    scope_parts.extend(self.current_function_scope)
                prefix = '.'.join(filter(None, scope_parts))
                if prefix:
                    var_name = f"{prefix}.{nm}"
                else:
                    var_name = nm

            self.add_def(var_name, "variable", node.lineno)
            if self.current_function_scope and self.local_var_maps:
                self.local_var_maps[-1][nm] = var_name
        else:
            self.visit(node.target)
        self.visit(node.value)

    def visit_Subscript(self, node):
        self.visit(node.value)
        self.visit(node.slice)

    def visit_Slice(self, node):
        if node.lower:
            self.visit(node.lower)
        if node.upper:
            self.visit(node.upper)
        if node.step:
            self.visit(node.step)

    def visit_Assign(self, node):
        def process_target_for_def(target_node):
            if isinstance(target_node, ast.Name):
                name_simple = target_node.id
                if (name_simple == "METADATA_DEPENDENCIES"
                    and self.cls and self.in_cst_class > 0):
                    return
                if name_simple == "__all__" and not self.current_function_scope and not self.cls:
                    return

                scope_parts = [self.mod]
                if self.cls:
                    scope_parts.append(self.cls)
                if self.current_function_scope:
                    scope_parts.extend(self.current_function_scope)

                if (self.current_function_scope and self.local_var_maps 
                        and name_simple in self.local_var_maps[-1]):
                    var_name = self.local_var_maps[-1][name_simple]
                else:
                    prefix = '.'.join(filter(None, scope_parts))
                    if prefix:
                        var_name = f"{prefix}.{name_simple}"
                    else:
                        var_name = name_simple

                self.add_def(var_name, "variable", target_node.lineno)
                if (self.current_function_scope and self.local_var_maps 
                        and name_simple not in self.local_var_maps[-1]):
                    self.local_var_maps[-1][name_simple] = var_name

            elif isinstance(target_node, (ast.Tuple, ast.List)):
                for elt in target_node.elts:
                    process_target_for_def(elt)

        for t in node.targets:
            process_target_for_def(t)

        for target in node.targets:
            if isinstance(target, ast.Name) and target.id == "__all__":
                if isinstance(node.value, (ast.List, ast.Tuple)):
                    for elt in node.value.elts:
                        value = None
                        if isinstance(elt, ast.Constant) and isinstance(elt.value, str):
                            value = elt.value
                        elif hasattr(elt, 's') and isinstance(elt.s, str):
                            value = elt.s

                        if value is not None:
                            if self.mod:
                                export_name = f"{self.mod}.{value}"
                            else:
                                export_name = value

                            self.add_ref(export_name)
                            self.add_ref(value)

        try:
            if isinstance(node.value, ast.Call):
                callee = node.value.func
                fqname = None

                if isinstance(callee, ast.Name):
                    fqname = self.alias.get(callee.id, self.qual(callee.id))

                elif isinstance(callee, ast.Attribute):
                    parts = []
                    cur = callee
                    while isinstance(cur, ast.Attribute):
                        parts.append(cur.attr)
                        cur = cur.value
                    head = None
                    if isinstance(cur, ast.Name):
                        head = self.alias.get(cur.id, self.qual(cur.id))
                    if head:
                        fqname = ".".join([head] + list(reversed(parts)))

                if fqname and self.current_function_scope and self.local_type_maps:
                    def _mark_target(t):
                        if isinstance(t, ast.Name):
                            self.local_type_maps[-1][t.id] = fqname
                        elif isinstance(t, (ast.Tuple, ast.List)):
                            for elt in t.elts:
                                _mark_target(elt)
                    for t in node.targets:
                        _mark_target(t)
        except Exception:
            pass

        self.generic_visit(node)

    def visit_Call(self, node):
        self.generic_visit(node)
        
        if isinstance(node.func, ast.Name) and node.func.id in ("getattr", "hasattr") and len(node.args) >= 2:
            if isinstance(node.args[1], ast.Constant) and isinstance(node.args[1].value, str):
                attr_name = node.args[1].value
                self.add_ref(attr_name)
                
                if isinstance(node.args[0], ast.Name):
                    module_name = node.args[0].id
                    if module_name != "self": 
                        qualified_name = f"{self.qual(module_name)}.{attr_name}"
                        self.add_ref(qualified_name)
            
            elif isinstance(node.args[0], ast.Name):
                target_name = node.args[0].id
                if target_name != "self":
                    if self.mod:
                        self.dyn.add(self.mod.split(".")[0])
                    else:
                        self.dyn.add("")
        
        elif isinstance(node.func, ast.Name) and node.func.id == "globals":
            parent = getattr(node, 'parent', None)
            if (isinstance(parent, ast.Subscript) and 
                isinstance(parent.slice, ast.Constant) and 
                isinstance(parent.slice.value, str)):
                func_name = parent.slice.value
                self.add_ref(func_name)
                self.add_ref(f"{self.mod}.{func_name}")
        
        elif isinstance(node.func, ast.Name) and node.func.id in ("eval", "exec"):
            root_mod = ""
            if self.mod:
                root_mod = self.mod.split(".")[0]
            self.dyn.add(root_mod)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            for param_name, param_full_name in self.current_function_params:
                if node.id == param_name:
                    self.first_read_lineno.setdefault(param_full_name, node.lineno)
                    self.add_ref(param_full_name)
                    return
            
            if (self.current_function_scope and self.local_var_maps
                and node.id in self.local_var_maps[-1]):
                fq = self.local_var_maps[-1][node.id]
                self.first_read_lineno.setdefault(fq, node.lineno)
                self.add_ref(fq)
                return
            
            qualified = self.qual(node.id)
            self.first_read_lineno.setdefault(qualified, node.lineno)
            self.add_ref(qualified)
            if node.id in DYNAMIC_PATTERNS:
                self.dyn.add(self.mod.split(".")[0])

    def visit_Attribute(self, node):
        self.generic_visit(node)

        if not isinstance(node.ctx, ast.Load):
            return

        if isinstance(node.value, ast.Name):
            base = node.value.id

            param_hit = None
            for param_name, param_full in self.current_function_params:
                if base == param_name:
                    param_hit = (param_name, param_full)
                    break

            if param_hit:
                self.add_ref(param_hit[1])

            if self.cls and base in {"self", "cls"}:
                if self.mod:
                    owner = f"{self.mod}.{self.cls}"
                else:
                    owner = self.cls

                self.add_ref(f"{owner}.{node.attr}")
                return

            if (self.current_function_scope and 
                self.local_type_maps and 
                self.local_type_maps[-1].get(base)):
                
                self.add_ref(f"{self.local_type_maps[-1][base]}.{node.attr}")
                return

            self.add_ref(f"{self.qual(base)}.{node.attr}")

    def visit_NamedExpr(self, node):
        self.visit(node.value)
        if isinstance(node.target, ast.Name):
            nm = node.target.id
            scope_parts = [self.mod]
            if self.cls: scope_parts.append(self.cls)
            if self.current_function_scope: scope_parts.extend(self.current_function_scope)
            prefix = '.'.join(filter(None, scope_parts))
            var_name = f"{prefix}.{nm}" if prefix else nm
            self.add_def(var_name, "variable", node.lineno)
            if self.current_function_scope and self.local_var_maps:
                self.local_var_maps[-1][nm] = var_name
            self.add_ref(var_name)

    def visit_keyword(self, node):
        self.visit(node.value)

    def visit_withitem(self, node):
        self.visit(node.context_expr)
        if node.optional_vars:
            self.visit(node.optional_vars)

    def visit_ExceptHandler(self, node):
        if node.type:
            self.visit(node.type)
        for stmt in node.body:
            self.visit(stmt)

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        item.parent = node
                        self.visit(item)
            elif isinstance(value, ast.AST):
                value.parent = node
                self.visit(value)


import pytest
import json
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch
from collections import defaultdict
from skylos.test_aware import TestAwareVisitor
from skylos.framework_aware import FrameworkAwareVisitor

from skylos.analyzer import (
    Skylos,  
    proc_file, 
    analyze
)

@pytest.fixture
def mock_definition():
    def _create_mock_def(name, simple_name, type, references=0, is_exported=False, confidence=100, in_init=False, line=1):
        mock = Mock()
        mock.name = name
        mock.simple_name = simple_name
        mock.type = type
        mock.references = references
        mock.is_exported = is_exported
        mock.confidence = confidence
        mock.in_init = in_init
        mock.line = line
        mock.to_dict.return_value = {
            "name": name,
            "type": type,
            "file": "test.py",
            "line": line
        }
        return mock
    return _create_mock_def

@pytest.fixture
def mock_test_aware_visitor():
    mock = Mock(spec=TestAwareVisitor)
    mock.is_test_file = False
    mock.test_decorated_lines = set()
    return mock

@pytest.fixture
def mock_framework_aware_visitor():
    mock = Mock(spec=FrameworkAwareVisitor)
    mock.framework_decorated_lines = set()
    return mock

@pytest.fixture
def temp_python_project():
    """Create a temp Python project for testing"""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        
        main_py = temp_path / "main.py"
        main_py.write_text("""
def used_function():
    return "used"

def unused_function():
    return "unused"

class UsedClass:
    def method(self):
        pass

class UnusedClass:
    def method(self):
        pass

result = used_function()
instance = UsedClass()
""")
        
        package_dir = temp_path / "mypackage"
        package_dir.mkdir()
        
        init_py = package_dir / "__init__.py"
        init_py.write_text("""
from .module import exported_function

def internal_function():
    pass
""")
        
        module_py = package_dir / "module.py"
        module_py.write_text("""
def exported_function():
    return "exported"

def internal_function():
    return "internal"
""")
        
        yield temp_path

class TestSkylos:
    
    @pytest.fixture
    def skylos(self):
        return Skylos()
    
    def test_init(self, skylos):
        assert skylos.defs == {}
        assert skylos.refs == []
        assert skylos.dynamic == set()
        assert isinstance(skylos.exports, defaultdict)
    
    def test_module_name_generation(self, skylos):
        root = Path("/project")
        
        # test a regular Python file
        file_path = Path("/project/src/module.py")
        result = skylos._module(root, file_path)
        assert result == "src.module"
        
        # test __init__.py file
        file_path = Path("/project/src/__init__.py")
        result = skylos._module(root, file_path)
        assert result == "src"
        
        # nested module
        file_path = Path("/project/src/package/submodule.py")
        result = skylos._module(root, file_path)
        assert result == "src.package.submodule"
        
        # root level file
        file_path = Path("/project/main.py")
        result = skylos._module(root, file_path)
        assert result == "main"
    
    def test_should_exclude_file(self, skylos):
        """
        should exclude pycache, build, egg-info and whatever is in exclude_folders
        """
        root = Path("/project")
        exclude_folders = {"__pycache__", "build", "*.egg-info"}
        
        file_path = Path("/project/src/__pycache__/module.pyc")
        assert skylos._should_exclude_file(file_path, root, exclude_folders)
        
        file_path = Path("/project/build/lib/module.py")
        assert skylos._should_exclude_file(file_path, root, exclude_folders)
        
        file_path = Path("/project/mypackage.egg-info/PKG-INFO")
        assert skylos._should_exclude_file(file_path, root, exclude_folders)
        
        file_path = Path("/project/src/module.py")
        assert not skylos._should_exclude_file(file_path, root, exclude_folders)
        
        assert not skylos._should_exclude_file(file_path, root, None)
    
    @patch('skylos.analyzer.Path')
    def test_get_python_files_single_file(self, mock_path, skylos):
        mock_file = Mock()
        mock_file.is_file.return_value = True
        mock_file.parent = Path("/project")
        mock_path.return_value.resolve.return_value = mock_file
        
        files, root = skylos._get_python_files("/project/test.py")
        assert files == [mock_file]
        assert root == Path("/project")
    
    @patch('skylos.analyzer.Path')
    def test_get_python_files_directory(self, mock_path, skylos):
        mock_dir = Mock()
        mock_dir.is_file.return_value = False
        mock_files = [Path("/project/file1.py"), Path("/project/file2.py")]
        mock_dir.glob.return_value = mock_files
        mock_path.return_value.resolve.return_value = mock_dir
        
        files, root = skylos._get_python_files("/project")
        assert files == mock_files
        assert root == mock_dir
    
    def test_mark_exports_in_init(self, skylos):
        mock_def1 = Mock()
        mock_def1.in_init = True
        mock_def1.simple_name = "public_function"
        mock_def1.is_exported = False
        
        mock_def2 = Mock()
        mock_def2.in_init = True
        mock_def2.simple_name = "_private_function"
        mock_def2.is_exported = False
        
        skylos.defs = {
            "module.public_function": mock_def1,
            "module._private_function": mock_def2
        }
        
        skylos._mark_exports()
        
        assert mock_def1.is_exported == True
        assert mock_def2.is_exported == False
    
    def test_mark_exports_explicit_exports(self, skylos):
        mock_def = Mock()
        mock_def.simple_name = "my_function"
        mock_def.type = "function"
        mock_def.is_exported = False
        
        skylos.defs = {"module.my_function": mock_def}
        skylos.exports = {"module": {"my_function"}}
        
        skylos._mark_exports()
        
        assert mock_def.is_exported == True
    
    def test_mark_refs_direct_reference(self, skylos):
        mock_def = Mock()
        mock_def.references = 0
        
        skylos.defs = {"module.function": mock_def}
        skylos.refs = [("module.function", None)]
        
        skylos._mark_refs()
        
        assert mock_def.references == 1
    
    def test_mark_refs_import_reference(self, skylos):
        mock_import = Mock()
        mock_import.type = "import"
        mock_import.simple_name = "imported_func"
        mock_import.references = 0
        
        mock_original = Mock()
        mock_original.type = "function"
        mock_original.simple_name = "imported_func"
        mock_original.references = 0
        
        skylos.defs = {
            "module.imported_func": mock_import,
            "other_module.imported_func": mock_original
        }
        skylos.refs = [("module.imported_func", None)]
        
        skylos._mark_refs()
        
        assert mock_import.references == 1
        assert mock_original.references == 1

class TestHeuristics:
    
    @pytest.fixture
    def skylos_with_class_methods(self, mock_definition):
        skylos = Skylos()
        
        mock_class = mock_definition(
            name="MyClass", 
            simple_name="MyClass", 
            type="class", 
            references=1
        )
        
        mock_init = mock_definition(
            name="MyClass.__init__",
            simple_name="__init__",
            type="method",
            references=0
        )
        
        mock_enter = mock_definition(
            name="MyClass.__enter__",
            simple_name="__enter__",
            type="method",
            references=0
        )
        
        skylos.defs = {
            "MyClass": mock_class,
            "MyClass.__init__": mock_init,
            "MyClass.__enter__": mock_enter
        }
        
        return skylos, mock_class, mock_init, mock_enter
    
    def test_auto_called_methods_get_references(self, skylos_with_class_methods):
        """auto-called methods get reference counts when class is used."""
        skylos, _, mock_init, mock_enter = skylos_with_class_methods
        
        skylos._apply_heuristics()
        
        assert mock_init.references == 1
        assert mock_enter.references == 1

class TestAnalyze:
    
    @patch('skylos.analyzer.proc_file')
    def test_analyze_basic(self, mock_proc_file, temp_python_project):
        mock_def = Mock()
        mock_def.name = "test.unused_function"
        mock_def.references = 0
        mock_def.is_exported = False
        mock_def.confidence = 80
        mock_def.type = "function"
        mock_def.to_dict.return_value = {
            "name": "test.unused_function",
            "type": "function", 
            "file": "test.py",
            "line": 1
        }
        
        mock_test_visitor = Mock(spec=TestAwareVisitor)
        mock_test_visitor.is_test_file = False
        mock_test_visitor.test_decorated_lines = set()
        
        mock_framework_visitor = Mock(spec=FrameworkAwareVisitor)
        mock_framework_visitor.framework_decorated_lines = set()
        mock_framework_visitor.is_framework_file = False
        
        mock_proc_file.return_value = ([mock_def], [], set(), set(), mock_test_visitor, mock_framework_visitor)
        
        result_json = analyze(str(temp_python_project), conf=60)
        result = json.loads(result_json)
        
        assert "unused_functions" in result
        assert "unused_imports" in result  
        assert "unused_classes" in result
        assert "unused_variables" in result
        assert "unused_parameters" in result
        assert "analysis_summary" in result
    
    def test_analyze_with_exclusions(self, temp_python_project):
        """analyze with folder exclusions."""
        exclude_dir = temp_python_project / "build"
        exclude_dir.mkdir()
        exclude_file = exclude_dir / "generated.py"
        exclude_file.write_text("def generated_function(): pass")
        
        result_json = analyze(str(temp_python_project), exclude_folders=["build"])
        result = json.loads(result_json)
        
        assert result["analysis_summary"]["excluded_folders"] == ["build"]
    
    def test_analyze_empty_directory(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            result_json = analyze(temp_dir, conf=60)
            result = json.loads(result_json)
            
            assert result["analysis_summary"]["total_files"] == 0
            assert all(len(result[key]) == 0 for key in [
                "unused_functions", "unused_imports", "unused_classes",
                "unused_variables", "unused_parameters"
            ])
    
    def test_confidence_threshold_filtering(self, mock_definition):
        """confidence threshold properly filters results."""
        skylos = Skylos()
        
        high_conf = mock_definition(
            name="high_conf",
            simple_name="high_conf",
            type="function",
            references=0,
            is_exported=False,
            confidence=80
        )
        
        low_conf = mock_definition(
            name="low_conf",
            simple_name="low_conf",
            type="function",
            references=0,
            is_exported=False,
            confidence=40
        )
        
        skylos.defs = {"high_conf": high_conf, "low_conf": low_conf}
        
        with patch.object(skylos, '_get_python_files') as mock_get_files:
            mock_get_files.return_value = ([Path("/fake/file.py")], Path("/"))
            
            with patch('skylos.analyzer.proc_file') as mock_proc_file:
                mock_proc_file.return_value = ([], [], set(), set(), Mock(spec=TestAwareVisitor), Mock(spec=FrameworkAwareVisitor))
                
                result_json = skylos.analyze("/fake/path", thr=60)
                result = json.loads(result_json)
                
                # include only high confidence 
                assert len(result["unused_functions"]) == 1
                assert result["unused_functions"][0]["name"] == "high_conf"

class TestProcFile:
    
    def test_proc_file_with_valid_python(self):
        with tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False) as f:
            f.write("""
def test_function():
    pass

class TestClass:
    def method(self):
        pass
""")
            f.flush()
            
            try:
                with patch('skylos.analyzer.Visitor') as mock_visitor_class, \
                     patch('skylos.analyzer.TestAwareVisitor') as mock_test_visitor_class, \
                     patch('skylos.analyzer.FrameworkAwareVisitor') as mock_framework_visitor_class:
                    
                    mock_visitor = Mock()
                    mock_visitor.defs = []
                    mock_visitor.refs = []
                    mock_visitor.dyn = set()
                    mock_visitor.exports = set()
                    mock_visitor_class.return_value = mock_visitor
                    
                    mock_test_visitor = Mock(spec=TestAwareVisitor)
                    mock_test_visitor_class.return_value = mock_test_visitor
                    
                    mock_framework_visitor = Mock(spec=FrameworkAwareVisitor)
                    mock_framework_visitor_class.return_value = mock_framework_visitor
                    
                    defs, refs, dyn, exports, test_flags, framework_flags = proc_file(f.name, "test_module")
                    
                    mock_visitor_class.assert_called_once_with("test_module", f.name)
                    mock_visitor.visit.assert_called_once()
                    
                    assert defs == []
                    assert refs == []
                    assert dyn == set()
                    ## added new test here for the new methods
                    assert exports == set()
                    assert test_flags == mock_test_visitor
                    assert framework_flags == mock_framework_visitor
            finally:
                Path(f.name).unlink()
    
    def test_proc_file_with_invalid_python(self):
        with tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False) as f:
            f.write("def invalid_syntax(:\npass") 
            f.flush()
            
            try:
                defs, refs, dyn, exports, test_flags, framework_flags = proc_file(f.name, "test_module")
                
                assert defs == []
                assert refs == []
                assert dyn == set()
                assert exports == set()
                assert isinstance(test_flags, TestAwareVisitor)
                assert isinstance(framework_flags, FrameworkAwareVisitor)
            finally:
                Path(f.name).unlink()
    
    def test_proc_file_with_tuple_args(self):
        with tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False) as f:
            f.write("def test(): pass")
            f.flush()
            
            try:
                with patch('skylos.analyzer.Visitor') as mock_visitor_class, \
                     patch('skylos.analyzer.TestAwareVisitor') as mock_test_visitor_class, \
                     patch('skylos.analyzer.FrameworkAwareVisitor') as mock_framework_visitor_class:
                    
                    mock_visitor = Mock()
                    mock_visitor.defs = []
                    mock_visitor.refs = []
                    mock_visitor.dyn = set()
                    mock_visitor.exports = set()
                    mock_visitor_class.return_value = mock_visitor
                    
                    mock_test_visitor = Mock(spec=TestAwareVisitor)
                    mock_test_visitor_class.return_value = mock_test_visitor
                    
                    mock_framework_visitor = Mock(spec=FrameworkAwareVisitor)
                    mock_framework_visitor_class.return_value = mock_framework_visitor
                    
                    defs, refs, dyn, exports, test_flags, framework_flags = proc_file((f.name, "test_module"))
                    
                    mock_visitor_class.assert_called_once_with("test_module", f.name)
            finally:
                Path(f.name).unlink()

class TestApplyPenalties:
    
    @patch('skylos.analyzer.detect_framework_usage')
    def test_private_name_penalty(self, mock_detect_framework, mock_definition, mock_test_aware_visitor, mock_framework_aware_visitor):
        """private names get penalized."""
        mock_detect_framework.return_value = None  # or whatever confidence value here that can change later on
        
        skylos = Skylos()
        mock_def = mock_definition(
            name="_private_func",
            simple_name="_private_func", 
            type="function",
            confidence=100
        )
        
        skylos._apply_penalties(mock_def, mock_test_aware_visitor, mock_framework_aware_visitor)
        assert mock_def.confidence < 100

    @patch('skylos.analyzer.detect_framework_usage')
    def test_magic_methods_confidence_zero(self, mock_detect_framework, mock_definition, mock_test_aware_visitor, mock_framework_aware_visitor):
        """magic methods get confidence of 0."""
        mock_detect_framework.return_value = None
        skylos = Skylos()
        mock_def = mock_definition(
            name="MyClass.__str__",
            simple_name="__str__",
            type="method",
            confidence=100
        )
        
        skylos._apply_penalties(mock_def, mock_test_aware_visitor, mock_framework_aware_visitor)
        assert mock_def.confidence == 0
    
    @patch('skylos.analyzer.detect_framework_usage') 
    def test_self_cls_parameters_confidence_zero(self, mock_detect_framework, mock_definition, mock_test_aware_visitor, mock_framework_aware_visitor):
        mock_detect_framework.return_value = None
        skylos = Skylos()
        
        mock_self = mock_definition(
            name="MyClass.method.self",
            simple_name="self",
            type="parameter",
            confidence=100
        )
        
        mock_cls = mock_definition(
            name="MyClass.classmethod.cls",
            simple_name="cls", 
            type="parameter",
            confidence=100
        )
        
        skylos._apply_penalties(mock_self, mock_test_aware_visitor, mock_framework_aware_visitor)
        skylos._apply_penalties(mock_cls, mock_test_aware_visitor, mock_framework_aware_visitor)
        
        assert mock_self.confidence == 0
        assert mock_cls.confidence == 0
    
    @patch('skylos.analyzer.detect_framework_usage')
    def test_test_methods_confidence_zero(self, mock_detect_framework, mock_definition, mock_framework_aware_visitor):
        """test methods get confidence of 0"""
        mock_detect_framework.return_value = None

        skylos = Skylos()
        
        test_visitor = Mock(spec=TestAwareVisitor)
        test_visitor.is_test_file = True
        test_visitor.test_decorated_lines = set()
        
        mock_def = mock_definition(
            name="TestMyClass.test_something",
            simple_name="test_something",
            type="method",
            confidence=100
        )
        
        skylos._apply_penalties(mock_def, test_visitor, mock_framework_aware_visitor)
        assert mock_def.confidence == 0
    
    @patch('skylos.analyzer.detect_framework_usage')
    def test_underscore_variable_confidence_zero(self, mock_detect_framework, mock_definition, mock_test_aware_visitor, mock_framework_aware_visitor):
        """underscore variables get confidence of 0."""
        mock_detect_framework.return_value = None

        skylos = Skylos()
        
        mock_def = mock_definition(
            name="_",
            simple_name="_",
            type="variable",
            confidence=100
        )
        
        skylos._apply_penalties(mock_def, mock_test_aware_visitor, mock_framework_aware_visitor)
        assert mock_def.confidence == 0

class TestIgnorePragmas:
    def test_analyze_respects_ignore_pragmas(self, tmp_path):
        src = tmp_path / "demo.py"
        src.write_text(
            """
def used():
    pass

def unused_no_ignore():
    pass

def unused_ignore():   # pragma: no skylos
    pass

used()
"""
        )

        result_json = analyze(str(tmp_path), conf=0)
        result = json.loads(result_json)

        # collect names of functions flagged as unreachable
        unreachable = {item["name"].split(".")[-1] for item in result["unused_functions"]}

        # expectations
        assert "unused_no_ignore" in unreachable
        assert "unused_ignore" not in unreachable
        assert "used" not in unreachable


if __name__ == "__main__":
    pytest.main([__file__, "-v"])


import re

PENALTIES = {
    "private_name": 80,   
    "dunder_or_magic": 100, 
    "underscored_var": 100, 
    "in_init_file": 15, 
    "dynamic_module": 40,
    "test_related": 100,
    "framework_magic": 40,
}

TEST_FILE_RE = re.compile(r"(?:^|[/\\])tests?[/\\]|_test\.py$", re.I)
TEST_IMPORT_RE = re.compile(r"^(pytest|unittest|nose|mock|responses)(\.|$)")
TEST_DECOR_RE = re.compile(r"""^(
    pytest\.(fixture|mark) |
    patch(\.|$) |
    responses\.activate |
    freeze_time
)$""", re.X)

AUTO_CALLED = {"__init__", "__enter__", "__exit__"}
TEST_METHOD_PATTERN = re.compile(r"^test_\w+$")

UNITTEST_LIFECYCLE_METHODS = {
    'setUp', 'tearDown', 'setUpClass', 'tearDownClass', 
    'setUpModule', 'tearDownModule'
}

FRAMEWORK_FILE_RE = re.compile(r"(?:views|handlers|endpoints|routes|api)\.py$", re.I)

DEFAULT_EXCLUDE_FOLDERS = {
    "__pycache__", ".git", ".pytest_cache", ".mypy_cache", ".tox",
    "htmlcov", ".coverage", "build", "dist", "*.egg-info", "venv", ".venv"
}

def is_test_path(p):
    return bool(TEST_FILE_RE.search(str(p)))

def is_framework_path(p):
    return bool(FRAMEWORK_FILE_RE.search(str(p)))

def parse_exclude_folders(user_exclude_folders= None, use_defaults= True, include_folders= None):
    exclude_folders = set()
    
    if use_defaults:
        exclude_folders.update(DEFAULT_EXCLUDE_FOLDERS)
    
    if user_exclude_folders:
        exclude_folders.update(user_exclude_folders)
    
    if include_folders:
        for folder in include_folders:
            exclude_folders.discard(folder)
    
    return exclude_folders 


import argparse
import json
import sys
import logging
from skylos.constants import parse_exclude_folders, DEFAULT_EXCLUDE_FOLDERS
from skylos.server import start_server
from skylos.analyzer import analyze as run_analyze
from skylos.codemods import (
    remove_unused_import_cst,
    remove_unused_function_cst,
    comment_out_unused_import_cst,
    comment_out_unused_function_cst, 
)
import pathlib
import skylos

try:
    import inquirer
    INTERACTIVE_AVAILABLE = True
except ImportError:
    INTERACTIVE_AVAILABLE = False

class Colors:
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    MAGENTA = '\033[95m'
    CYAN = '\033[96m'
    BOLD = '\033[1m'
    RESET = '\033[0m'
    GRAY = '\033[90m'

class CleanFormatter(logging.Formatter):
    def format(self, record):
        return record.getMessage()

def setup_logger(output_file=None):
    logger = logging.getLogger('skylos')
    logger.setLevel(logging.INFO)
    
    logger.handlers.clear()
    
    formatter = CleanFormatter()
    
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    
    if output_file:
        file_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        file_handler = logging.FileHandler(output_file)
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)
    
    logger.propagate = False
    
    return logger

def remove_unused_import(file_path, import_name, line_number):
    path = pathlib.Path(file_path)
    try:
        src = path.read_text(encoding="utf-8")
        new_code, changed = remove_unused_import_cst(src, import_name, line_number)
        if not changed:
            return False
        path.write_text(new_code, encoding="utf-8")
        return True
    except Exception as e:
        logging.error(f"Failed to remove import {import_name} from {file_path}: {e}")
        return False

def remove_unused_function(file_path, function_name, line_number):
    path = pathlib.Path(file_path)
    try:
        src = path.read_text(encoding="utf-8")
        new_code, changed = remove_unused_function_cst(src, function_name, line_number)
        if not changed:
            return False
        path.write_text(new_code, encoding="utf-8")
        return True
    except Exception as e:
        logging.error(f"Failed to remove function {function_name} from {file_path}: {e}")
        return False
    
def comment_out_unused_import(file_path, import_name, line_number, marker="SKYLOS DEADCODE"):
    path = pathlib.Path(file_path)
    try:
        src = path.read_text(encoding="utf-8")
        new_code, changed = comment_out_unused_import_cst(src, import_name, line_number, marker=marker)
        if not changed:
            return False
        path.write_text(new_code, encoding="utf-8")
        return True
    except Exception as e:
        logging.error(f"Failed to comment out import {import_name} from {file_path}: {e}")
        return False

def comment_out_unused_function(file_path, function_name, line_number, marker="SKYLOS DEADCODE"):
    path = pathlib.Path(file_path)
    try:
        src = path.read_text(encoding="utf-8")
        new_code, changed = comment_out_unused_function_cst(src, function_name, line_number, marker=marker)
        if not changed:
            return False
        path.write_text(new_code, encoding="utf-8")
        return True
    except Exception as e:
        logging.error(f"Failed to comment out function {function_name} from {file_path}: {e}")
        return False

def interactive_selection(logger, unused_functions, unused_imports):
    if not INTERACTIVE_AVAILABLE:
        logger.error("Interactive mode requires 'inquirer' package. Install with: pip install inquirer")
        return [], []
    
    selected_functions = []
    selected_imports = []
    
    if unused_functions:
        logger.info(f"\n{Colors.CYAN}{Colors.BOLD}Select unused functions to remove (hit spacebar to select):{Colors.RESET}")
        
        function_choices = []

        for item in unused_functions:
            choice_text = f"{item['name']} ({item['file']}:{item['line']})"
            function_choices.append((choice_text, item))
        
        questions = [
            inquirer.Checkbox('functions',
                            message="Select functions to remove",
                            choices=function_choices,
                            )
        ]
        
        answers = inquirer.prompt(questions)
        if answers:
            selected_functions = answers['functions']
    
    if unused_imports:
        logger.info(f"\n{Colors.MAGENTA}{Colors.BOLD}Select unused imports to act on (hit spacebar to select):{Colors.RESET}")
        
        import_choices = []

        for item in unused_imports:
            choice_text = f"{item['name']} ({item['file']}:{item['line']})"
            import_choices.append((choice_text, item))
        
        questions = [
            inquirer.Checkbox('imports',
                            message="Select imports to remove",
                            choices=import_choices,
                            )
        ]
        
        answers = inquirer.prompt(questions)
        if answers:
            selected_imports = answers['imports']
    
    return selected_functions, selected_imports

def print_badge(dead_code_count: int, logger):
    logger.info(f"\n{Colors.GRAY}{'─' * 50}{Colors.RESET}")
    
    if dead_code_count == 0:
        logger.info(f" Your code is 100% dead code free! Add this badge to your README:")
        logger.info("```markdown")
        logger.info("![Dead Code Free](https://img.shields.io/badge/Dead_Code-Free-brightgreen?logo=moleculer&logoColor=white)")
        logger.info("```")
    else:
        logger.info(f"Found {dead_code_count} dead code items. Add this badge to your README:")
        logger.info("```markdown")  
        logger.info(f"![Dead Code: {dead_code_count}](https://img.shields.io/badge/Dead_Code-{dead_code_count}_detected-orange?logo=codacy&logoColor=red)")
        logger.info("```")

def main():
    if len(sys.argv) > 1 and sys.argv[1] == 'run':
        try:
            start_server()
            return
        except ImportError:
            print(f"{Colors.RED}Error: Flask is required {Colors.RESET}")
            print(f"{Colors.YELLOW}Install with: pip install flask flask-cors{Colors.RESET}")
            sys.exit(1)

    parser = argparse.ArgumentParser(
        description="Detect unreachable functions and unused imports in a Python project"
    )
    parser.add_argument("path", help="Path to the Python project")

    parser.add_argument(
        "--version",
        action="version", 
        version=f"skylos {skylos.__version__}",
        help="Show version and exit"
    )
    
    parser.add_argument(
        "--json",
        action="store_true",
        help="Output raw JSON",
    )

    parser.add_argument(
        "--comment-out",
        action="store_true",
        help="Comment out selected dead code instead of deleting it",
    )

    parser.add_argument(
        "--output",
        "-o",
        type=str,
        help="Write output to file",
    )
    parser.add_argument(
        "--verbose", "-v",
        action="store_true",
        help="Enable verbose"
    )
    parser.add_argument(
        "--confidence",
        "-c",
        type=int,
        default=60,
        help="Confidence threshold (0-100). Lower values include more items. Default: 60"
    )
    parser.add_argument(
        "--interactive", "-i",
        action="store_true",
        help="Select items to remove"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show what would be removed"
    )
    
    parser.add_argument(
        "--exclude-folder",
        action="append",
        dest="exclude_folders",
        help="Exclude a folder from analysis (can be used multiple times). "
             "By default, common folders like __pycache__, .git, venv are excluded. "
             "Use --no-default-excludes to disable default exclusions."
    )
    
    parser.add_argument(
        "--include-folder", 
        action="append",
        dest="include_folders",
        help="Force include a folder that would otherwise be excluded "
             "(overrides both default and custom exclusions). "
             "Example: --include-folder venv"
    )
    
    parser.add_argument(
        "--no-default-excludes",
        action="store_true",
        help="Don't exclude default folders (__pycache__, .git, venv, etc.). "
             "Only exclude folders with --exclude-folder."
    )
    
    parser.add_argument(
        "--list-default-excludes",
        action="store_true", 
        help="List the default excluded folders and exit."
    )

    parser.add_argument("--secrets", action="store_true",
                   help="Scan for API keys. Off by default.")
    
    args = parser.parse_args()

    if args.list_default_excludes:
        print("Default excluded folders:")
        for folder in sorted(DEFAULT_EXCLUDE_FOLDERS):
            print(f" {folder}")
        print(f"\nTotal: {len(DEFAULT_EXCLUDE_FOLDERS)} folders")
        print("\nUse --no-default-excludes to disable these exclusions")
        print("Use --include-folder <folder> to force include specific folders")
        return
    
    logger = setup_logger(args.output)
    
    if args.verbose:
        logger.setLevel(logging.DEBUG)
        logger.debug(f"Analyzing path: {args.path}")
        
        if args.exclude_folders:
            logger.debug(f"Excluding folders: {args.exclude_folders}")

    use_defaults = not args.no_default_excludes
    final_exclude_folders = parse_exclude_folders(
        user_exclude_folders=args.exclude_folders,
        use_defaults=use_defaults,
        include_folders=args.include_folders
    )
    
    if not args.json:
        if final_exclude_folders:
            logger.info(f"{Colors.YELLOW}📁 Excluding: {', '.join(sorted(final_exclude_folders))}{Colors.RESET}")
        else:
            logger.info(f"{Colors.GREEN}📁 No folders excluded{Colors.RESET}")

    try:
        result_json = run_analyze(args.path, conf=args.confidence, enable_secrets=bool(args.secrets), exclude_folders=list(final_exclude_folders))
        result = json.loads(result_json)

    except Exception as e:
        logger.error(f"Error during analysis: {e}")
        sys.exit(1)

    if args.json:
        logger.info(result_json)
        return

    unused_functions = result.get("unused_functions", [])
    unused_imports = result.get("unused_imports", [])
    unused_parameters = result.get("unused_parameters", [])
    unused_variables = result.get("unused_variables", [])
    unused_classes = result.get("unused_classes", [])
    secrets_findings = result.get("secrets", [])
    
    logger.info(f"{Colors.CYAN}{Colors.BOLD} Python Static Analysis Results{Colors.RESET}")
    logger.info(f"{Colors.CYAN}{'=' * 35}{Colors.RESET}")
    
    logger.info(f"\n{Colors.BOLD}Summary:{Colors.RESET}")
    logger.info(f" * Unreachable functions: {Colors.YELLOW}{len(unused_functions)}{Colors.RESET}")
    logger.info(f" * Unused imports: {Colors.YELLOW}{len(unused_imports)}{Colors.RESET}")
    logger.info(f" * Unused parameters: {Colors.YELLOW}{len(unused_parameters)}{Colors.RESET}")
    logger.info(f" * Unused variables: {Colors.YELLOW}{len(unused_variables)}{Colors.RESET}")
    logger.info(f" * Unused classes: {Colors.YELLOW}{len(unused_classes)}{Colors.RESET}")
    if secrets_findings:
        logger.info(f" * Secrets: {Colors.RED}{len(secrets_findings)}{Colors.RESET}")

    if args.interactive and (unused_functions or unused_imports):
        logger.info(f"\n{Colors.BOLD}Interactive Mode:{Colors.RESET}")
        selected_functions, selected_imports = interactive_selection(logger, unused_functions, unused_imports)
        
        if selected_functions or selected_imports:
            logger.info(f"\n{Colors.BOLD}Selected items to process:{Colors.RESET}")
            
            if selected_functions:
                logger.info(f" Functions: {len(selected_functions)}")
                for func in selected_functions:
                    logger.info(f"  - {func['name']} ({func['file']}: {func['line']})")
            
            if selected_imports:
                logger.info(f" Imports: {len(selected_imports)}")
                for imp in selected_imports:
                    logger.info(f"  - {imp['name']} ({imp['file']}: {imp['line']})")
            
            if not args.dry_run:
                if args.comment_out:
                    confirm_verb = "comment out"
                else:
                    confirm_verb = "remove"

                questions = [
                    inquirer.Confirm('confirm',
                                   message="Are you sure you want to process these items?",
                                   default=False)
                ]
                answers = inquirer.prompt(questions)
                
                if answers and answers['confirm']:
                    action = "Commenting out" if args.comment_out else "Removing"
                    logger.info(f"\n{Colors.YELLOW}{action} selected items...{Colors.RESET}")

                    action_func = comment_out_unused_function if args.comment_out else remove_unused_function
                    if args.comment_out:
                        action_past = "Commented out"
                        action_verb = "comment out"
                    else:
                        action_past = "Removed"
                        action_verb = "remove"

                    for func in selected_functions:
                        success = action_func(func['file'], func['name'], func['line'])
                        
                        if success:
                            logger.info(f"  {Colors.GREEN} ✓ {Colors.RESET} {action_past} function: {func['name']}")
                        else:
                            logger.error(f"  {Colors.RED} x {Colors.RESET} Failed to {action_verb}: {func['name']}")
                            
                    import_func = comment_out_unused_import if args.comment_out else remove_unused_import
                    if args.comment_out:
                        action_past = "Commented out"
                        action_verb = "comment out"
                    else:
                        action_past = "Removed"
                        action_verb = "remove"

                    for imp in selected_imports:
                        success = import_func(imp['file'], imp['name'], imp['line'])
                        
                        if success:
                            logger.info(f"  {Colors.GREEN} ✓ {Colors.RESET} {action_past} import: {imp['name']}")
                        else:
                            logger.error(f"  {Colors.RED} x {Colors.RESET} Failed to {action_verb}: {imp['name']}")
                            
                    logger.info(f"\n{Colors.GREEN}Cleanup complete!{Colors.RESET}")
                else:
                    logger.info(f"\n{Colors.YELLOW}Operation cancelled.{Colors.RESET}")
            else:
                logger.info(f"\n{Colors.YELLOW}Dry run - no files were modified.{Colors.RESET}")
        else:
            logger.info(f"\n{Colors.BLUE}No items selected.{Colors.RESET}")
    
    else:
        if unused_functions:
            logger.info(f"\n{Colors.RED}{Colors.BOLD} - Unreachable Functions{Colors.RESET}")
            logger.info(f"{Colors.RED}{'=' * 23}{Colors.RESET}")
            for i, item in enumerate(unused_functions, 1):
                logger.info(f"{Colors.GRAY}{i:2d}. {Colors.RESET}{Colors.RED}{item['name']}{Colors.RESET}")
                logger.info(f"    {Colors.GRAY}└─ {item['file']}:{item['line']}{Colors.RESET}")
        else:
            logger.info(f"\n{Colors.GREEN} All functions are reachable!{Colors.RESET}")
        
        if unused_imports:
            logger.info(f"\n{Colors.MAGENTA}{Colors.BOLD} - Unused Imports{Colors.RESET}")
            logger.info(f"{Colors.MAGENTA}{'=' * 16}{Colors.RESET}")
            for i, item in enumerate(unused_imports, 1):
                logger.info(f"{Colors.GRAY}{i:2d}. {Colors.RESET}{Colors.MAGENTA}{item['name']}{Colors.RESET}")
                logger.info(f"    {Colors.GRAY}└─ {item['file']}:{item['line']}{Colors.RESET}")
        else:
            logger.info(f"\n{Colors.GREEN}✓ All imports are being used!{Colors.RESET}")
        
        if unused_parameters:
            logger.info(f"\n{Colors.BLUE}{Colors.BOLD} - Unused Parameters{Colors.RESET}")
            logger.info(f"{Colors.BLUE}{'=' * 18}{Colors.RESET}")
            for i, item in enumerate(unused_parameters, 1):
                logger.info(f"{Colors.GRAY}{i:2d}. {Colors.RESET}{Colors.BLUE}{item['name']}{Colors.RESET}")
                logger.info(f"    {Colors.GRAY}└─ {item['file']}:{item['line']}{Colors.RESET}")
        else:
            logger.info(f"\n{Colors.GREEN}✓ All parameters are being used!{Colors.RESET}")
        
        if unused_variables:
            logger.info(f"\n{Colors.YELLOW}{Colors.BOLD} - Unused Variables{Colors.RESET}")
            logger.info(f"{Colors.YELLOW}{'=' * 18}{Colors.RESET}")
            for i, item in enumerate(unused_variables, 1):
                logger.info(f"{Colors.GRAY}{i:2d}. {Colors.RESET}{Colors.YELLOW}{item['name']}{Colors.RESET}")
                logger.info(f"    {Colors.GRAY}└─ {item['file']}:{item['line']}{Colors.RESET}")
        else:
            logger.info(f"\n{Colors.GREEN}✓ All variables are being used!{Colors.RESET}")
                
        if unused_classes:
            logger.info(f"\n{Colors.YELLOW}{Colors.BOLD} - Unused Classes{Colors.RESET}")
            logger.info(f"{Colors.YELLOW}{'=' * 18}{Colors.RESET}")
            for i, item in enumerate(unused_classes, 1):
                logger.info(f"{Colors.GRAY}{i:2d}. {Colors.RESET}{Colors.YELLOW}{item['name']}{Colors.RESET}")
                logger.info(f"    {Colors.GRAY}└─ {item['file']}:{item['line']}{Colors.RESET}")
        else:
            logger.info(f"\n{Colors.GREEN}✓ All classes are being used!{Colors.RESET}")

        if secrets_findings:
            logger.info(f"\n{Colors.RED}{Colors.BOLD} - Secrets{Colors.RESET}")
            logger.info(f"{Colors.RED}{'=' * 9}{Colors.RESET}")
            for i, s in enumerate(secrets_findings[:20], 1):
                provider = s.get("provider", "generic")
                where = f"{s.get('file','?')}:{s.get('line','?')}"
                prev = s.get("preview", "****")
                msg = s.get("message", "Secret detected")
                logger.info(f"{Colors.GRAY}{i:2d}. {Colors.RESET}{msg} [{provider}] {Colors.GRAY}({where}){Colors.RESET} -> {prev}")

        dead_code_count = len(unused_functions) + len(unused_imports) + len(unused_variables) + len(unused_classes) + len(unused_parameters)

        print_badge(dead_code_count, logger)

        if unused_functions or unused_imports:
            logger.info(f"\n{Colors.BOLD}Next steps:{Colors.RESET}")
            logger.info(f" * Use --select specific items to remove")
            logger.info(f" * Use --dry-run to preview changes")
            logger.info(f" * Use --exclude-folder to skip directories")

if __name__ == "__main__":
    main()


from __future__ import annotations
import libcst as cst
from libcst.metadata import PositionProvider
from libcst.helpers import get_full_name_for_node

class _CommentOutBlock(cst.CSTTransformer):

    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, module_code, marker = "SKYLOS DEADCODE"):
        self.module_code = module_code.splitlines(True)
        self.marker = marker

    def _comment_block(self, start_line, end_line):
        lines = self.module_code[start_line - 1:end_line]
        out = []
        out.append(cst.EmptyLine(comment=cst.Comment(f"# {self.marker} START (lines {start_line}-{end_line})")))
        for raw in lines:
            out.append(cst.EmptyLine(comment=cst.Comment("# " + raw.rstrip("\n"))))
        out.append(cst.EmptyLine(comment=cst.Comment(f"# {self.marker} END")))
        return out

class _CommentOutFunctionAtLine(_CommentOutBlock):
    def __init__(self, func_name, target_line, module_code, marker):
        super().__init__(module_code, marker)
        self.func_name = func_name
        self.target_line = target_line
        self.changed = False

    def _is_target(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return pos and pos.start.line == self.target_line

    def leave_FunctionDef(self, orig: cst.FunctionDef, updated: cst.FunctionDef):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            pos = self.get_metadata(PositionProvider, orig)
            return cst.FlattenSentinel(self._comment_block(pos.start.line, pos.end.line))
        return updated

    def leave_AsyncFunctionDef(self, orig: cst.AsyncFunctionDef, updated: cst.AsyncFunctionDef):
        target = self.func_name.split(".")[-1] 
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            pos = self.get_metadata(PositionProvider, orig)
            return cst.FlattenSentinel(self._comment_block(pos.start.line, pos.end.line))
        return updated

class _CommentOutImportAtLine(_CommentOutBlock):

    def __init__(self, target_name, target_line, module_code, marker):
        super().__init__(module_code, marker)
        self.target_name = target_name
        self.target_line = target_line
        self.changed = False

    def _is_target_line(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and (pos.start.line <= self.target_line <= pos.end.line))

    def _render_single_alias_text(self, head, alias: cst.ImportAlias, is_from):
        if is_from:
            alias_txt = alias.name.code
            if alias.asname:
                alias_txt += f" as {alias.asname.name.value}"
            return f"from {head} import {alias_txt}"
        else:
            alias_txt = alias.name.code
            if alias.asname:
                alias_txt += f" as {alias.asname.name.value}"
            return f"import {alias_txt}"

    def _split_aliases(self, aliases, head, is_from):
        kept = []
        removed_for_comment= []
        for alias in list(aliases):
            bound = _bound_name_for_import_alias(alias)
            name_code = get_full_name_for_node(alias.name)
            tail = name_code.split(".")[-1]
            if self.target_name in (bound, tail):
                self.changed = True
                removed_for_comment.append(self._render_single_alias_text(head, alias, is_from))
            else:
                kept.append(alias)
        return kept, removed_for_comment

    def leave_Import(self, orig: cst.Import, updated: cst.Import):
        if not self._is_target_line(orig):
            return updated
        
        head = "" 
        kept, removed = self._split_aliases(updated.names, head, is_from=False)
        
        if not removed:
            return updated
        
        pos = self.get_metadata(PositionProvider, orig)
        if not kept:
            return cst.FlattenSentinel(self._comment_block(pos.start.line, pos.end.line))
        
        commented = []
        for txt in removed:
            comment = cst.Comment(f"# {self.marker}: {txt}")
            commented.append(cst.EmptyLine(comment=comment))

        kept_import = updated.with_changes(names=tuple(kept))
        all_nodes = [kept_import] + commented
        return cst.FlattenSentinel(all_nodes)

    def leave_ImportFrom(self, orig: cst.ImportFrom, updated: cst.ImportFrom):
        if not self._is_target_line(orig) or isinstance(updated.names, cst.ImportStar):
            return updated
        
        if updated.relative:
            dots = "." * len(updated.relative)
        else:
            dots = ""

        if updated.module is not None:
            modname = updated.module.code
        else:
            modname = ""

        mod = f"{dots}{modname}"

        kept, removed = self._split_aliases(list(updated.names), mod, is_from=True)
        
        if not removed:
            return updated
        pos = self.get_metadata(PositionProvider, orig)
        
        if not kept:
            comment_block = self._comment_block(pos.start.line, pos.end.line)
            return cst.FlattenSentinel(comment_block)

        commented = []
        for txt in removed:
            comment = cst.Comment(f"# {self.marker}: {txt}")
            commented.append(cst.EmptyLine(comment=comment))

        updated_import = updated.with_changes(names=tuple(kept))
        all_nodes = [updated_import] + commented

        return cst.FlattenSentinel(all_nodes)

def comment_out_unused_function_cst(code, func_name, line_number, marker = "SKYLOS DEADCODE"):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _CommentOutFunctionAtLine(func_name, line_number, code, marker)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed

def comment_out_unused_import_cst(code, import_name, line_number, marker = "SKYLOS DEADCODE"):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _CommentOutImportAtLine(import_name, line_number, code, marker)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed

def _bound_name_for_import_alias(alias: cst.ImportAlias):
    if alias.asname:
        return alias.asname.name.value
    node = alias.name
    while isinstance(node, cst.Attribute):
        node = node.value
    return node.value 

class _RemoveImportAtLine(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, target_name, target_line):
        self.target_name = target_name
        self.target_line = target_line
        self.changed = False

    def _is_target_line(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and (pos.start.line <= self.target_line <= pos.end.line))
    
    def _filter_aliases(self, aliases):
        kept = []
        for alias in aliases:
            bound = _bound_name_for_import_alias(alias)
            name_code = get_full_name_for_node(alias.name) or ""
            tail = name_code.split(".")[-1]
            if self.target_name in (bound, tail):
                self.changed = True
                continue
            kept.append(alias)
        return kept

    def leave_Import(self, orig: cst.Import, updated: cst.Import):
        if not self._is_target_line(orig):
            return updated
        kept = self._filter_aliases(updated.names)
        if not kept:
            return cst.RemoveFromParent()
        return updated.with_changes(names=tuple(kept))

    def leave_ImportFrom(self, orig: cst.ImportFrom, updated: cst.ImportFrom):
        if not self._is_target_line(orig):
            return updated
        if isinstance(updated.names, cst.ImportStar):
            return updated 
        kept = self._filter_aliases(list(updated.names))
        if not kept:
            return cst.RemoveFromParent()

        return updated.with_changes(names=tuple(kept))

class _RemoveFunctionAtLine(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, func_name, target_line):
        self.func_name = func_name
        self.target_line = target_line
        self.changed = False

    def _is_target(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return pos and pos.start.line == self.target_line

    def leave_FunctionDef(self, orig: cst.FunctionDef, updated: cst.FunctionDef):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            return cst.RemoveFromParent()
        return updated

    def leave_AsyncFunctionDef(self, orig: cst.AsyncFunctionDef, updated: cst.AsyncFunctionDef):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            return cst.RemoveFromParent()

        return updated

def remove_unused_import_cst(code, import_name, line_number):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _RemoveImportAtLine(import_name, line_number)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed

def remove_unused_function_cst(code, func_name, line_number):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _RemoveFunctionAtLine(func_name, line_number)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed
